mod jsonrpc;
mod transform;
mod transports;
use std::path::Path;

use futures::future::try_join;
use futures::prelude::*;
use irc::client::prelude::*;
use irc::client::ClientStream;
use jsonrpsee::core::client::{Subscription, SubscriptionClientT};
use serde_json::Value;
use transform::Bridge;
use {log, pretty_env_logger};

use crate::jsonrpc::RpcClient;

const CHANNEL: &str = "#hackens-signal";
const MEDIA_DIR: &str = "./www";
const SIGNALCLI_DIR: &str = "./.config";
const SOCKET: &str = "./socket";
const URL_ROOT: &str = "https://hackens.org/irc-bridge/";

#[tokio::main]
async fn main() -> Result<(), String> {
    pretty_env_logger::init();
    log::info!("Starting...");
    // TODO: load config at runtime and add consts inside
    let config = Config {
        nickname: Some("bottest".to_owned()),
        server: Some("ulminfo.fr".to_owned()),
        channels: vec![CHANNEL.to_owned()],
        use_tls: Some(true),
        ..Config::default()
    };

    // Initialisation of components. TODO separate module
    let init_irc = async {
        let mut client = Client::from_config(config).await.unwrap();
        client.identify().unwrap();

        let sender = client.sender();
        let mut stream = client.stream().unwrap();

        log::debug!("Initializing IRC");
        while client.list_channels().is_some_and(|x| x.len() == 0) {
            stream
                .next()
                .await
                .transpose()
                .map_err(|e| format!("error waiting for motd: {e}"))?;
        }
        log::debug!("Initialized IRC");
        Ok::<(irc::client::Sender, ClientStream), String>((sender, stream))
    };
    let init_signal = async {
        log::debug!("Initializing Signal");

        let signal_client = jsonrpc::connect_unix(SOCKET)
            .await
            .map_err(|e| format!("failed to connect to signal socket: {e}"))?;
        let signal_stream = signal_client.subscribe_receive(None).await.unwrap();
        log::debug!("Initialized Signal");
        Ok((signal_stream, signal_client))
    };
    let ((sender, mut stream), (mut signal_stream, signal_client)) =
        try_join(init_irc, init_signal).await?;
    let bridge = Bridge::new(
        &Path::new(MEDIA_DIR),
        &Path::new(SIGNALCLI_DIR),
        URL_ROOT,
        GROUPID,
    );

    // Run !!
    log::info!("Bridge is up and running");

    try_join(
        handle_signal(sender, &mut signal_stream, &bridge),
        handle_irc(&mut stream, signal_client, &bridge),
    )
    .await?;
    unreachable!();
}

async fn handle_irc(
    stream: &mut ClientStream,
    signal_client: impl SubscriptionClientT + std::marker::Sync,
    bridge: &Bridge<'_>,
) -> Result<(), String> {
    loop {
        if let Some(message) = stream
            .next()
            .await
            .transpose()
            .map_err(|e| format!("error while retrieving messages from irc: {e}"))?
        {
            if let Some(msg) = bridge.irc2signal(message) {
                log::trace!("[SIGNAL SEND] {msg}");
                signal_client
                    .send(
                        None,
                        vec![],
                        vec![String::from(GROUPID)],
                        false,
                        false,
                        msg,
                        vec![],
                        vec![],
                        vec![],
                        None,
                        None,
                        None,
                        vec![],
                        vec![],
                        vec![],
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                    )
                    .await
                    .map_err(|e| format!("error while sending to signal: {e}"))?;
            }
        }
    }
}

async fn handle_signal(
    sender: Sender,
    stream: &mut Subscription<Value>,
    bridge: &Bridge<'_>,
) -> Result<(), String> {
    loop {
        match stream.next().await {
            None => {
                return Err(String::from("signal-cli stream terminated"));
            }
            Some(Ok(val)) => {
                log::trace!("[SIGNAL RCV] {:?}", val);
                if let Some(msg) = bridge.signal2irc(val).await {
                    sender
                        .send_privmsg(CHANNEL, msg)
                        .map_err(|e| format!("error while sending to irc: {e}"))?;
                }
            }
            Some(Err(err)) => {
                return Err(format!(
                    "error while retrieving messages from signal: {err}"
                ))
            }
        }
    }
}
