/*
 * This work is adapted from signal-cli (https://github.com/AsamK/signal-cli/)
 * Copyright (C) 2024 AsamK
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
use futures_util::stream::StreamExt;
use futures_util::{Sink, SinkExt, Stream};
use jsonrpsee::core::async_trait;
use jsonrpsee::core::client::{ReceivedMessage, TransportReceiverT, TransportSenderT};
use thiserror::Error;

pub mod ipc;
mod stream_codec;

#[derive(Debug, Error)]
enum Errors {
    #[error("Other: {0}")]
    Other(String),
    #[error("Closed")]
    Closed,
}

struct Sender<T: Send + Sink<String>> {
    inner: T,
}

#[async_trait]
impl<T: Send + Sink<String, Error = impl std::error::Error> + Unpin + 'static> TransportSenderT
    for Sender<T>
{
    type Error = Errors;

    async fn send(&mut self, body: String) -> Result<(), Self::Error> {
        self.inner
            .send(body)
            .await
            .map_err(|e| Errors::Other(format!("{:?}", e)))?;
        Ok(())
    }

    async fn close(&mut self) -> Result<(), Self::Error> {
        self.inner
            .close()
            .await
            .map_err(|e| Errors::Other(format!("{:?}", e)))?;
        Ok(())
    }
}

struct Receiver<T: Send + Stream> {
    inner: T,
}

#[async_trait]
impl<T: Send + Stream<Item = Result<String, std::io::Error>> + Unpin + 'static> TransportReceiverT
    for Receiver<T>
{
    type Error = Errors;

    async fn receive(&mut self) -> Result<ReceivedMessage, Self::Error> {
        match self.inner.next().await {
            None => Err(Errors::Closed),
            Some(Ok(msg)) => Ok(ReceivedMessage::Text(msg)),
            Some(Err(e)) => Err(Errors::Other(format!("{:?}", e))),
        }
    }
}
