/*
 * This work is adapted from signal-cli (https://github.com/AsamK/signal-cli/)
 * Copyright (C) 2024 AsamK
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
use std::{io, str};

use bytes::BytesMut;
use tokio_util::codec::{Decoder, Encoder};

type Separator = u8;

/// Stream codec for streaming protocols (ipc, tcp)
#[derive(Debug, Default)]
pub struct StreamCodec {
    incoming_separator: Separator,
    outgoing_separator: Separator,
}

impl StreamCodec {
    /// Default codec with streaming input data. Input can be both enveloped and not.
    pub fn stream_incoming() -> Self {
        StreamCodec::new(b'\n', b'\n')
    }

    /// New custom stream codec
    pub fn new(incoming_separator: Separator, outgoing_separator: Separator) -> Self {
        StreamCodec {
            incoming_separator,
            outgoing_separator,
        }
    }
}

impl Decoder for StreamCodec {
    type Item = String;
    type Error = io::Error;

    fn decode(&mut self, buf: &mut BytesMut) -> io::Result<Option<Self::Item>> {
        if let Some(i) = buf
            .as_ref()
            .iter()
            .position(|&b| b == self.incoming_separator)
        {
            let line = buf.split_to(i);
            let _ = buf.split_to(1);

            match str::from_utf8(line.as_ref()) {
                Ok(s) => Ok(Some(s.to_string())),
                Err(_) => Err(io::Error::new(io::ErrorKind::Other, "invalid UTF-8")),
            }
        } else {
            Ok(None)
        }
    }
}

impl Encoder<String> for StreamCodec {
    type Error = io::Error;

    fn encode(&mut self, msg: String, buf: &mut BytesMut) -> io::Result<()> {
        let mut payload = msg.into_bytes();
        payload.push(self.outgoing_separator);
        buf.extend_from_slice(&payload);
        Ok(())
    }
}
